/*
 * Interrupt handling.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef CONFIG_CPU_JZ4730_MINIPC
#include "minipc.h"
#else
#include "nanonote.h"
#endif

#include "board-specific.h"
#include "board.h"
#include "lcd.h"
#include "jzlcd.h"
#include "cpu.h"
#include "irq.h"

extern vidinfo_t panel_info;

void next_pixel(unsigned short *x, unsigned short *y)
{
	(*x)++;
	if (*x >= panel_info.vl_col) {
		*x = 0;
		(*y)++;
		if (*y >= panel_info.vl_row)
			*y = 0;
	}
}

/* Task management. */

enum { max_tasks = 3 };
static u32 stack_pointers[max_tasks];
u8 current_task = 0;
u32 current_stack_pointer;
extern u32 _got_copy_start;

const u32 stack_start = 0x00080000;
const u32 stack_size = 0x00002000;
const u32 pagesize = 4 * 1024;
const u32 framesize = 120; /* see the handlers */

/* Tasks. */

void plot_pattern(unsigned short pixel_type, unsigned short x, unsigned short y)
{
	while (1) {
		if (pixel_type)
			test_pixel(x, y);
		else
			clear_pixel(x, y);
		next_pixel(&x, &y);
		udelay(100);
	}
}

/* Initialisation and handling. */

void irq_init()
{
	handle_error_level();
	timer_init_irq();
	init_interrupts();
}

void irq_handle()
{
	unsigned short i;

	/* Check interrupt identity. */

	if (REG_INTC_IPR & (1 << TIMER_CHAN_IRQ)) {

		/* Switch task. */

		/* switch_task(); */

		/* Clear interrupt status. */

		__intc_ack_irq(TIMER_CHAN_IRQ);
		__tcu_clear_full_match_flag(TIMER_CHAN);

	/* Handle other interrupts, anyway. */

	} else {
		for (i = 0; i < 32; i++) {
			if (REG_INTC_IPR & (1 << i))
				__intc_ack_irq(i);
		}
	}
}

void start_task(unsigned short task)
{
	u32 args[] = {1, 0, (task - 1) * 120};
	u32 virtual, physical;

	/*
	Each task employs a stack at a multiple of the given start address in
	physical memory, but at the same address in virtual memory.
	*/

	virtual = stack_start - pagesize * 2;
	physical = stack_start + stack_size * task - pagesize * 2;

	init_page_table(page_table_start, virtual, physical, pagesize, 0x1e, task);

	/* Map the page for initialisation. */

	map_page(physical, physical, pagesize, 0x1e, 0);

	/*
	Set the stack for the new task, initialising the global pointer and
	return address.
	*/

	init_stack(stack_start + stack_size * task, _got_copy_start, (void (*)()) plot_pattern, args, 3);

	/* Advance the stack pointer so that the adopted frame will be found. */

	stack_pointers[task] = stack_start - framesize;
}

void switch_task()
{
	/*
	Get the current stack pointer. This is obtained just after saving the
	task's state.
	*/

	stack_pointers[current_task] = current_stack_pointer;

	/* Switch the current task. */

	current_task++;
	if (current_task == max_tasks) current_task = 1;

	/*
	Set the current stack pointer. This is actually set just before
	restoring the task's state.
	*/

	current_stack_pointer = stack_pointers[current_task];
}

void invoke_task(unsigned short task)
{
	current_task = task;
	current_stack_pointer = stack_pointers[current_task];
	set_task(current_task);

	asm volatile(
		".set noat\n"
		"move $sp, %0\n"
		"addi $sp, $sp, 120\n"
		"lw $at, -4($sp)\n"
		"lw $v0, -8($sp)\n"
		"lw $v1, -12($sp)\n"
		"lw $a0, -16($sp)\n"
		"lw $a1, -20($sp)\n"
		"lw $a2, -24($sp)\n"
		"lw $a3, -28($sp)\n"
		"lw $t0, -32($sp)\n"
		"lw $t1, -36($sp)\n"
		"lw $t2, -40($sp)\n"
		"lw $t3, -44($sp)\n"
		"lw $t4, -48($sp)\n"
		"lw $t5, -52($sp)\n"
		"lw $t6, -56($sp)\n"
		"lw $t7, -60($sp)\n"
		"lw $s0, -64($sp)\n"
		"lw $s1, -68($sp)\n"
		"lw $s2, -72($sp)\n"
		"lw $s3, -76($sp)\n"
		"lw $s4, -80($sp)\n"
		"lw $s5, -84($sp)\n"
		"lw $s6, -88($sp)\n"
		"lw $s7, -92($sp)\n"
		"lw $t8, -96($sp)\n"
		"lw $t9, -100($sp)\n"
		"lw $gp, -104($sp)\n"
		"lw $fp, -108($sp)\n"
		"lw $ra, -112($sp)\n"
		"lw $k0, -116($sp)\n"
		"mtc0 $k0, $14\n" /* CP0_EPC */
		"nop\n"
		"jr $ra\n"
		"nop"
		:
		: "r" (current_stack_pointer)
		);
}
